@@ -219,7 +219,7 @@ def read_excel(
 def read_excel(
     source: FileSource,
     *,
-    sheet_id: None,
+    sheet_id: None = ...,
     sheet_name: list[str] | tuple[str],
     table_name: str | None = ...,
     engine: ExcelSpreadsheetEngine = ...,
@@ -265,6 +265,10 @@ def read_excel(
         Support loading data from a list (or glob pattern) of multiple workbooks.
     .. versionchanged:: 1.0
         Default engine is now "calamine" (was "xlsx2csv").
+    .. versionchanged:: 0.20.7
+        The `read_csv_options` parameter was renamed `read_options`.
+    .. versionchanged:: 0.20.6
+        The `xlsx2csv_options` parameter was renamed `engine_options`.
 
     Parameters
     ----------
@@ -510,7 +514,7 @@ def read_ods(
 def read_ods(
     source: FileSource,
     *,
-    sheet_id: None,
+    sheet_id: None = ...,
     sheet_name: list[str] | tuple[str],
     has_header: bool = ...,
     columns: Sequence[int] | Sequence[str] | None = ...,
@@ -909,7 +913,7 @@ def _csv_buffer_to_frame(
     if schema_overrides:
         if csv_dtypes := read_options.get("dtypes", {}):
             issue_deprecation_warning(
-                "The `dtypes` parameter for `read_csv` is deprecated. It has been renamed to `schema_overrides`.",
+                "the `dtypes` parameter for `read_csv` is deprecated. It has been renamed to `schema_overrides`.",
                 version="0.20.31",
             )
 
